import pytest

from apiron import Endpoint, Service, ServiceBase


@pytest.fixture
def service():
    class SomeService(Service):
        domain = "http://foo.com"

    return SomeService


class TestServiceBase:
    def test_get_hosts_returns_empty_list_by_default(self):
        assert ServiceBase.get_hosts() == []

    def test_required_headers_returns_empty_dict_by_default(self, service):
        assert service.required_headers == {}


class TestService:
    def test_get_hosts_returns_domain(self, service):
        assert service.get_hosts() == ["http://foo.com"]

    def test_str_method_on_class(self, service):
        assert str(service) == "http://foo.com"

    def test_repr_method_on_class(self, service):
        assert repr(service) == "SomeService(domain=http://foo.com)"

    def test_required_headers_returns_empty_dict_by_default(self, service):
        assert service.required_headers == {}

    def test_endpoints_when_no_endpoints(self, service):
        assert service.endpoints == set()

    def test_endpoints_when_one_endpoint(self, service):
        foo = Endpoint(path="/foo")
        service.foo = foo
        assert service.endpoints == {foo}

    def test_endpoints_when_multiple_endpoints(self, service):
        foo = Endpoint(path="/foo")
        bar = Endpoint(path="/bar")

        service.foo = foo
        service.bar = bar

        assert service.endpoints == {foo, bar}



import pytest
import requests


@pytest.fixture(autouse=True)
def _hobble_network(monkeypatch, request):
    """Hobble all network calls made through the requests library"""

    if "no_hobble_network" in request.keywords:
        return

    def hobbled_send(session, request_object, **kwargs):
        raise RuntimeError(f"requests hobbled for testing: tried calling {request_object.url}")

    monkeypatch.setattr(requests.Session, "send", hobbled_send)



from __future__ import annotations

import collections
import logging
import random
from typing import TYPE_CHECKING, Any
from urllib import parse

import requests
from requests import adapters
from urllib3.util import retry

if TYPE_CHECKING:
    import apiron  # pragma: no cover

from apiron.exceptions import NoHostsAvailableException

LOGGER = logging.getLogger(__name__)

DEFAULT_CONNECTION_TIMEOUT = 1
DEFAULT_READ_TIMEOUT = 3

DEFAULT_CONNECTION_RETRIES = 1
DEFAULT_READ_RETRIES = 1
DEFAULT_TOTAL_RETRIES = 1
DEFAULT_STATUS_CODES_TO_RETRY_ON = range(500, 600)

Timeout = collections.namedtuple("Timeout", ["connection_timeout", "read_timeout"])

DEFAULT_TIMEOUT = Timeout(connection_timeout=DEFAULT_CONNECTION_TIMEOUT, read_timeout=DEFAULT_READ_TIMEOUT)
DEFAULT_RETRY = retry.Retry(
    total=DEFAULT_TOTAL_RETRIES,
    connect=DEFAULT_CONNECTION_RETRIES,
    read=DEFAULT_READ_RETRIES,
    status_forcelist=DEFAULT_STATUS_CODES_TO_RETRY_ON,
)


def _build_url(host: str, path: str) -> str:
    """
    Builds a valid URL from a host and path which may or may not have slashes in the proper place.
    Does not conform to `IETF RFC 1808 <https://tools.ietf.org/html/rfc1808.html>`_ but instead joins the host and path as given.
    Does not append any additional slashes to the final URL; just joins the host and path properly.

    :param str host:
        An HTTP host like ``'https://awesome-api.com/v2'``
    :param str path:
        The path to an endpoint on the host like ``'/some-resource/'``
    :return:
        The properly-joined URL of host and path, e.g. ``'https://awesome-api.com/v2/some-resource/'``
    :rtype:
        str
    """
    host += "/" if not host.endswith("/") else ""
    path = path.lstrip("/")

    return parse.urljoin(host, path)


def _adapt_session(session: requests.Session, adapter: requests.adapters.HTTPAdapter) -> requests.Session:
    """
    Mounts an adapter capable of communication over HTTP or HTTPS to the supplied session.

    :param adapter:
        A :class:`requests.adapters.HTTPAdapter` instance
    :return:
        The adapted :class:`requests.Session` instance
    """
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def _get_required_headers(service: apiron.Service, endpoint: apiron.Endpoint) -> dict[str, str]:
    """
    :param Service service:
        The service being called
    :param Endpoint endpoint:
        The endpoint being called
    :return:
        Headers required by the ``service`` and the ``endpoint`` being called
    :rtype:
        dict
    """
    headers = {}
    headers.update(service.required_headers)
    headers.update(endpoint.required_headers)
    return headers


def _choose_host(service: apiron.Service) -> str:
    hosts = service.get_hosts()
    if not hosts:
        raise NoHostsAvailableException(getattr(service, "service_name", "UNKNOWN SERVICE"))
    return random.choice(hosts)  # noqa: S311


def _build_request_object(
    session: requests.Session,
    service: apiron.Service,
    endpoint: apiron.Endpoint,
    method: str | None = None,
    params: dict[str, Any] | None = None,
    data: dict[str, Any] | None = None,
    files: dict[str, str] | None = None,
    json: dict[str, Any] | None = None,
    headers: dict[str, Any] | None = None,
    cookies: dict[str, Any] | None = None,
    auth: Any | None = None,
    **kwargs,
):
    host = _choose_host(service=service)

    path = endpoint.get_formatted_path(**kwargs)

    merged_params = endpoint.get_merged_params(params)

    headers = headers or {}
    headers.update(_get_required_headers(service, endpoint))

    request = requests.Request(
        method=method or endpoint.default_method,
        url=_build_url(host, path),
        params=merged_params,
        data=data,
        files=files,
        json=json,
        headers=headers,
        cookies=cookies,
        auth=auth,
    )

    return session.prepare_request(request)


def _get_guaranteed_session(session: requests.Session | None) -> requests.Session:
    if session:
        return session
    return requests.Session()


def _get_retry_spec(endpoint: apiron.Endpoint, retry_spec: retry.Retry | None = None) -> retry.Retry:
    return retry_spec or endpoint.retry_spec or DEFAULT_RETRY


def _get_timeout_spec(endpoint: apiron.Endpoint, timeout_spec: Timeout | None = None) -> Timeout:
    return timeout_spec or endpoint.timeout_spec or DEFAULT_TIMEOUT


def call(
    service: apiron.Service,
    endpoint: apiron.Endpoint,
    method: str | None = None,
    session: requests.Session | None = None,
    params: dict[str, Any] | None = None,
    data: dict[str, Any] | None = None,
    files: dict[str, str] | None = None,
    json: dict[str, Any] | None = None,
    headers: dict[str, Any] | None = None,
    cookies: dict[str, Any] | None = None,
    auth: Any | None = None,
    encoding: str | None = None,
    retry_spec: retry.Retry | None = None,
    timeout_spec: Timeout | None = None,
    logger: logging.Logger | None = None,
    allow_redirects: bool = True,
    return_raw_response_object: bool | None = None,
    **kwargs,
):
    """
    :param Service service:
        The service that hosts the endpoint being called
    :param Endpoint endpoint:
        The endpoint being called
    :param str method:
        The HTTP method to use for the call
    :param requests.Session session:
        (optional)
        An existing session, useful for making many calls in a single session
        (default ``None``)
    :param dict params:
        (optional)
        ``GET`` parameters to send to the endpoint
        (default ``None``)
    :param dict data:
        (optional)
        ``POST`` data to send to the endpoint.
        A :class:`dict` will be form-encoded, while a :class:`str` will be sent raw
        (default ``None``)
    :param dict files:
        (optional)
        Dictionary of ``'filename': file-like-objects`` for multipart encoding upload.
        (default ``None``)
    :param dict json:
        (optional)
        A JSON-serializable dictionary that will be sent as the ``POST`` body
        (default ``None``)
    :param dict headers:
        HTTP Headers to send to the endpoint
        (default ``None``)
    :param dict cookies:
        Cookies to send to the endpoint
        (default ``None``)
    :param auth:
        An object suitable for the :class:`requests.Request` object's ``auth`` argument
    :param str encoding:
        The codec to use when decoding the response.
        Default behavior is to have ``requests`` guess the codec.
        (default ``None``)
    :param urllib3.util.retry.Retry retry_spec:
        (optional)
        An override of the retry behavior for this call.
        (default ``None``)
    :param Timeout timeout_spec:
        (optional)
        An override of the timeout behavior for this call.
        (default ``None``)
    :param logging.Logger logger:
        (optional)
        An existing logger for logging from the proper caller for better correlation
    :param bool allow_redirects:
        (optional)
        Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection
        (default ``True``)
    :param bool return_raw_response_object:
        Whether to return a :class:`requests.Response` object or call :func:`format_response` on it first.
        (Default ``False``)
    :param ``**kwargs``:
        Arguments to be formatted into the ``endpoint`` argument's ``path`` attribute
    :return:
        The result of ``endpoint``'s :func:`format_response`
    :rtype: The type returned by ``endpoint``'s :func:`format_response`
    :raises requests.RetryError:
        if retry threshold exceeded due to bad HTTP codes (default 500 range)
    :raises requests.ConnectionError:
        if retry threshold exceeded due to connection or request timeouts
    """
    logger = logger or LOGGER

    managing_session = not session
    guaranteed_session = _get_guaranteed_session(session)

    retry_spec_to_use = _get_retry_spec(endpoint, retry_spec)

    adapted_session = _adapt_session(guaranteed_session, adapters.HTTPAdapter(max_retries=retry_spec_to_use))

    method = method or endpoint.default_method

    auth = auth or getattr(session, "auth", None) or service.auth

    request = _build_request_object(
        adapted_session,
        service,
        endpoint,
        method=method,
        params=params,
        data=data,
        files=files,
        json=json,
        headers=headers,
        cookies=cookies,
        auth=auth,
        **kwargs,
    )

    logger.info("%s %s", method, request.url)

    timeout_spec_to_use = _get_timeout_spec(endpoint, timeout_spec)

    response = adapted_session.send(
        request,
        timeout=(
            timeout_spec_to_use.connection_timeout,
            timeout_spec_to_use.read_timeout,
        ),
        stream=getattr(endpoint, "streaming", False),
        allow_redirects=allow_redirects,
        proxies=adapted_session.proxies or service.proxies,
    )

    logger.info(
        "%d %s%s",
        response.status_code,
        response.url,
        f" ({len(response.history)} redirect(s))" if response.history else "",
    )

    if managing_session:
        adapted_session.close()

    response.raise_for_status()

    if encoding:
        response.encoding = encoding

    # Use the explicitly passed in option, if any
    # Otherwise, use the endpoint's setting
    if return_raw_response_object is None:
        return_raw_response = endpoint.return_raw_response_object
    else:
        return_raw_response = return_raw_response_object

    return response if return_raw_response else endpoint.format_response(response)



from typing import Protocol

from apiron.service.base import ServiceBase


class Resolver(Protocol):
    @staticmethod
    def resolve(service_name: str) -> list[str]: ...


class DiscoverableService(ServiceBase):
    """
    A Service whose hosts are determined via a host resolver.
    A host resolver is any class with a :func:`resolve` method
    that takes a service name as its sole argument
    and returns a list of host names that correspond to that service.
    """

    host_resolver_class: type[Resolver]
    service_name: str

    @classmethod
    def get_hosts(cls) -> list[str]:
        return cls.host_resolver_class.resolve(cls.service_name)

    def __str__(self) -> str:
        return self.service_name

    def __repr__(self) -> str:
        klass = self.__class__
        return (
            f"{klass.__name__}(service_name={klass.service_name}, host_resolver={klass.host_resolver_class.__name__})"
        )



import collections
from collections.abc import Iterable
from typing import Any

from apiron.endpoint.endpoint import Endpoint


class JsonEndpoint(Endpoint):
    """
    An endpoint that returns :mimetype:`application/json`
    """

    def __init__(
        self,
        *args,
        path: str = "/",
        default_method: str = "GET",
        default_params: dict[str, Any] | None = None,
        required_params: Iterable[str] | None = None,
        preserve_order: bool = False,
    ):
        super().__init__(
            path=path,
            default_method=default_method,
            default_params=default_params,
            required_params=required_params,
        )
        self.preserve_order = preserve_order

    def format_response(self, response) -> dict[str, Any]:
        """
        Extracts JSON data from the response

        :param requests.Response response:
            The original response from :mod:`requests`
        :return:
            The response's JSON content
        :rtype:
            :class:`dict` if ``preserve_order`` is ``False``
        :rtype:
            :class:`collections.OrderedDict` if ``preserve_order`` is ``True``
        """

        return response.json(object_pairs_hook=collections.OrderedDict if self.preserve_order else None)

    @property
    def required_headers(self) -> dict[str, str]:
        return {"Accept": "application/json"}



from unittest import mock

import pytest
from urllib3.util import retry

from apiron import NoHostsAvailableException, Timeout, client


@pytest.fixture
def mock_response():
    response = mock.Mock()
    response.history = []
    return response


@pytest.fixture
def mock_endpoint():
    endpoint = mock.Mock()
    endpoint.required_headers = {}
    endpoint.get_formatted_path.return_value = "/foo/"
    endpoint.timeout_spec = None
    endpoint.retry_spec = None
    del endpoint.stub_response
    return endpoint


@pytest.fixture
def mock_logger():
    return mock.Mock()


@mock.patch("requests.sessions.Session", autospec=True)
def test_adapt_session(mock_session):
    adapter = mock.Mock()
    mock_session.get_adapter.return_value = adapter
    adapted_session = client._adapt_session(mock_session, adapter)
    assert adapter == adapted_session.get_adapter("http://foo.com")
    assert adapter == adapted_session.get_adapter("https://foo.com")


def test_get_required_headers(mock_endpoint):
    service = mock.Mock()
    service.required_headers = {"one": "two"}
    mock_endpoint.required_headers = {"foo": "bar"}
    expected_headers = {}
    expected_headers.update(service.required_headers)
    expected_headers.update(mock_endpoint.required_headers)
    assert expected_headers == client._get_required_headers(service, mock_endpoint)


@mock.patch("apiron.client.requests.Request")
@mock.patch("apiron.client._get_required_headers")
def test_build_request_object_passes_arguments_to_request_constructor(
    mock_get_required_headers, mock_request_constructor, mock_endpoint
):
    session = mock.Mock()

    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]

    mock_endpoint.default_method = "POST"
    mock_endpoint.required_headers = {"header": "value"}
    mock_endpoint.required_params = set()

    params = {"baz": "qux"}
    mock_endpoint.get_merged_params.return_value = params
    data = {"data": "I am a data"}
    files = {"file_name": "this is a test"}
    json = {"raw": "data"}
    headers = {"Accept": "stuff"}
    cookies = {"chocolate-chip": "yes"}
    auth = mock.Mock()

    mock_get_required_headers.return_value = {"header": "value"}
    expected_headers = {}
    expected_headers.update(headers)
    expected_headers.update(mock_endpoint.required_headers)

    with mock.patch.object(session, "prepare_request") as mock_prepare_request:
        client._build_request_object(
            session,
            service,
            mock_endpoint,
            params=params,
            data=data,
            files=files,
            json=json,
            headers=headers,
            cookies=cookies,
            auth=auth,
            foo="bar",
        )

        mock_request_constructor.assert_called_once_with(
            url="http://host1.biz/foo/",
            method=mock_endpoint.default_method,
            headers=expected_headers,
            cookies=cookies,
            params=params,
            data=data,
            files=files,
            json=json,
            auth=auth,
        )

        assert mock_prepare_request.call_count == 1


@mock.patch("apiron.client.Timeout")
@mock.patch("apiron.client._adapt_session")
@mock.patch("apiron.client._build_request_object")
@mock.patch("requests.adapters.HTTPAdapter", autospec=True)
@mock.patch("requests.Session", autospec=True)
def test_call(
    MockSession,
    MockAdapter,
    mock_build_request_object,
    mock_adapt_session,
    mock_timeout,
    mock_response,
    mock_endpoint,
    mock_logger,
):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]

    mock_endpoint.default_method = "GET"
    mock_endpoint.streaming = True

    request = mock.Mock()
    request.url = "http://host1.biz/foo/"
    mock_build_request_object.return_value = request

    mock_response.status_code = 200
    mock_response.url = "http://host1.biz/foo/"

    mock_session = MockSession()
    mock_session.send.return_value = mock_response
    mock_session.proxies = {}
    mock_session.auth = ()
    mock_adapt_session.return_value = mock_session

    client.call(service, mock_endpoint, timeout_spec=mock_timeout, logger=mock_logger)

    mock_adapt_session.assert_called_once_with(mock_session, MockAdapter())
    mock_session.send.assert_called_once_with(
        request,
        timeout=(mock_timeout.connection_timeout, mock_timeout.read_timeout),
        stream=mock_endpoint.streaming,
        allow_redirects=True,
        proxies=service.proxies,
    )

    mock_logger.info.assert_any_call("%s %s", "GET", "http://host1.biz/foo/")
    mock_logger.info.assert_any_call("%d %s%s", 200, "http://host1.biz/foo/", "")

    mock_endpoint.default_method = "POST"
    request.method = "POST"

    client.call(
        service,
        mock_endpoint,
        session=mock_session,
        timeout_spec=mock_timeout,
        logger=mock_logger,
    )

    mock_session.send.assert_any_call(
        request,
        timeout=(mock_timeout.connection_timeout, mock_timeout.read_timeout),
        stream=mock_endpoint.streaming,
        allow_redirects=True,
        proxies=service.proxies,
    )

    mock_logger.info.assert_any_call("%s %s", "GET", "http://host1.biz/foo/")
    mock_logger.info.assert_any_call("%d %s%s", 200, "http://host1.biz/foo/", "")

    request.method = "PUT"

    client.call(
        service,
        mock_endpoint,
        method="PUT",
        session=mock_session,
        timeout_spec=mock_timeout,
        logger=mock_logger,
    )

    mock_session.send.assert_any_call(
        request,
        timeout=(mock_timeout.connection_timeout, mock_timeout.read_timeout),
        stream=mock_endpoint.streaming,
        allow_redirects=True,
        proxies=service.proxies,
    )


@mock.patch("apiron.client._build_request_object")
@mock.patch("apiron.client._adapt_session")
@mock.patch("requests.Session", autospec=True)
def test_call_auth_priority(
    MockSession,
    mock_adapt_session,
    mock_build_request_object,
    mock_endpoint,
    mock_logger,
):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}
    service.auth = ("service-user", "p455w0rd!")

    mock_session = MockSession()
    mock_session.proxies = {}
    mock_session.auth = ("session-user", "p455w0rd!")

    mock_adapt_session.return_value = mock_session

    client.call(
        service,
        mock_endpoint,
        auth=("direct-user", "p455w0rd!"),
        session=mock_session,
        logger=mock_logger,
    )
    assert mock_build_request_object.call_args[1]["auth"] == (
        "direct-user",
        "p455w0rd!",
    )

    client.call(service, mock_endpoint, session=mock_session, logger=mock_logger)
    assert mock_build_request_object.call_args[1]["auth"] == (
        "session-user",
        "p455w0rd!",
    )

    mock_session.auth = ()
    client.call(service, mock_endpoint, logger=mock_logger)
    assert mock_build_request_object.call_args[1]["auth"] == (
        "service-user",
        "p455w0rd!",
    )


def test_call_with_existing_session(mock_response, mock_endpoint, mock_logger):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = mock.Mock()
    session.send.return_value = mock_response

    client.call(service, mock_endpoint, session=session, logger=mock_logger)

    assert not session.close.called


def test_call_with_explicit_encoding(mock_response, mock_endpoint, mock_logger):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = mock.Mock()
    session.send.return_value = mock_response

    client.call(
        service,
        mock_endpoint,
        session=session,
        logger=mock_logger,
        encoding="FAKE-CODEC",
    )

    assert mock_response.encoding == "FAKE-CODEC"


@mock.patch("apiron.client._build_request_object")
@mock.patch("apiron.client._adapt_session")
@mock.patch("requests.Session", autospec=True)
def test_call_uses_configured_endpoint_timeout_spec(
    MockSession,
    mock_adapt_session,
    mock_build_request_object,
    mock_response,
    mock_endpoint,
    mock_logger,
):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = MockSession()
    session.send.return_value = mock_response

    request = mock.Mock()
    mock_build_request_object.return_value = request
    mock_endpoint.timeout_spec = Timeout(connection_timeout=1_000, read_timeout=1_000)

    mock_session = MockSession()
    mock_session.send.return_value = mock_response
    mock_session.proxies = {}
    mock_session.auth = ()
    mock_adapt_session.return_value = mock_session

    client.call(service, mock_endpoint, session=mock_session, logger=mock_logger)

    session.send.assert_called_once_with(
        request,
        timeout=(1_000, 1_000),
        stream=mock_endpoint.streaming,
        allow_redirects=True,
        proxies=service.proxies,
    )


@mock.patch("apiron.client._build_request_object")
@mock.patch("apiron.client.adapters.HTTPAdapter", autospec=True)
@mock.patch("requests.Session", autospec=True)
def test_call_uses_configured_endpoint_retry_spec(
    MockSession,
    MockAdapter,
    mock_build_request_object,
    mock_response,
    mock_endpoint,
    mock_logger,
):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = MockSession()
    session.send.return_value = mock_response

    request = mock.Mock()
    mock_build_request_object.return_value = request
    retry_spec = retry.Retry(
        total=42,
        connect=42,
        read=42,
        status_forcelist=[500],
    )
    mock_endpoint.retry_spec = retry_spec

    mock_session = MockSession()
    mock_session.send.return_value = mock_response
    mock_session.proxies = {}
    mock_session.auth = ()

    client.call(service, mock_endpoint, session=mock_session, logger=mock_logger)

    MockAdapter.assert_called_once_with(max_retries=retry_spec)


def test_build_request_object_raises_no_host_exception():
    service = mock.Mock()
    service.get_hosts.return_value = []

    with pytest.raises(NoHostsAvailableException):
        client._build_request_object(mock.Mock(), service, mock.Mock())


def test_choose_host_returns_one_of_the_available_hosts():
    hosts = ["foo", "bar", "baz"]
    service = mock.Mock()
    service.get_hosts.return_value = hosts
    assert client._choose_host(service) in hosts


def test_choose_host_raises_exception_when_no_hosts_available():
    service = mock.Mock()
    service.get_hosts.return_value = []
    with pytest.raises(NoHostsAvailableException):
        client._choose_host(service)


def test_call_when_raw_response_object_requested(mock_response, mock_endpoint, mock_logger):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = mock.Mock()
    session.send.return_value = mock_response

    response = client.call(
        service,
        mock_endpoint,
        session=session,
        logger=mock_logger,
        return_raw_response_object=True,
    )

    assert response is mock_response


def test_call_when_raw_response_object_requested_on_endpoint(mock_response, mock_endpoint):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = mock.Mock()
    session.send.return_value = mock_response

    mock_endpoint.return_raw_response_object = True

    response = client.call(service, mock_endpoint, session=session, logger=mock.Mock())

    assert response is mock_response


def test_return_raw_response_object_in_call_overrides_endpoint(mock_response, mock_endpoint):
    service = mock.Mock()
    service.get_hosts.return_value = ["http://host1.biz"]
    service.required_headers = {}

    session = mock.Mock()
    session.send.return_value = mock_response

    mock_endpoint.return_raw_response_object = False

    response = client.call(
        service,
        mock_endpoint,
        session=session,
        logger=mock.Mock(),
        return_raw_response_object=True,
    )

    assert response is mock_response


@pytest.mark.parametrize(
    ("host", "path", "url"),
    [
        ("http://biz.com", "/endpoint", "http://biz.com/endpoint"),
        ("http://biz.com/", "endpoint", "http://biz.com/endpoint"),
        ("http://biz.com/", "/endpoint", "http://biz.com/endpoint"),
        ("http://biz.com", "endpoint", "http://biz.com/endpoint"),
        ("http://biz.com/v2", "/endpoint", "http://biz.com/v2/endpoint"),
        ("http://biz.com/v2/", "endpoint", "http://biz.com/v2/endpoint"),
        ("http://biz.com/v2/", "/endpoint", "http://biz.com/v2/endpoint"),
        ("http://biz.com/v2", "endpoint", "http://biz.com/v2/endpoint"),
    ],
)
def test_build_url(host, path, url):
    assert url == client._build_url(host, path)



from __future__ import annotations

import logging
import string
import warnings
from collections.abc import Callable, Iterable
from functools import partial, update_wrapper
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from typing import Concatenate, ParamSpec

    from apiron.service import Service

    P = ParamSpec("P")
    R = TypeVar("R")

import requests
from urllib3.util import retry

from apiron import Timeout, client
from apiron.exceptions import UnfulfilledParameterException

LOGGER = logging.getLogger(__name__)


def _create_caller(
    call_fn: Callable[Concatenate[Service, Endpoint, P], R],
    instance: Any,
    owner: Any,
) -> Callable[P, R]:
    return partial(call_fn, instance, owner)


class Endpoint:
    """
    A basic service endpoint that responds with the default ``Content-Type`` for that endpoint
    """

    def __get__(self, instance, owner):
        caller = _create_caller(client.call, owner, self)
        update_wrapper(caller, client.call)
        return caller

    def __call__(self):
        raise TypeError("Endpoints are only callable in conjunction with a Service class.")

    def __init__(
        self,
        path: str = "/",
        default_method: str = "GET",
        default_params: dict[str, Any] | None = None,
        required_params: Iterable[str] | None = None,
        return_raw_response_object: bool = False,
        timeout_spec: Timeout | None = None,
        retry_spec: retry.Retry | None = None,
    ):
        """
        :param str path:
            The URL path for this endpoint, without the protocol or domain
        :param str default_method:
            (Default ``'GET'``)
            The default method to use when calling this endpoint.
        :param dict default_params:
            The default parameters to use when calling this endpoint.
            Useful when an endpoint always or most often needs a base set of parameters supplied.
        :param required_params:
            An iterable of required parameter names.
            Calling an endpoint without its required parameters raises an exception.
        :param bool return_raw_response_object:
            Whether to return a :class:`requests.Response` object or call :func:`format_response` on it first.
            This can be overridden when calling the endpoint.
            (Default ``False``)
        :param Timeout timeout_spec:
            (optional)
            An override of the timeout behavior for calls to this endpoint.
            (default ``None``)
        :param urllib3.util.retry.Retry retry_spec:
            (optional)
            An override of the retry behavior for calls to this endpoint.
            (default ``None``)
        """
        self.default_method = default_method

        if "?" in path:
            warnings.warn(
                f"Endpoint path ('{path}') may contain query parameters. "
                f"Use the default_params or required_params attributes in the initialization of this endpoint, "
                f"or the params argument when calling the endpoint instead.",
                stacklevel=3,
            )

        self.path = path
        self.default_params = default_params or {}
        self.required_params = required_params or set()
        self.return_raw_response_object = return_raw_response_object
        self.timeout_spec = timeout_spec
        self.retry_spec = retry_spec

    def format_response(self, response: requests.Response) -> str | dict[str, Any] | Iterable[bytes]:
        """
        Extracts the appropriate type of response data from a :class:`requests.Response` object

        :param requests.Response response:
            The original response from :mod:`requests`
        :return:
            The response's text content
        :rtype:
            str
        """
        return response.text

    @property
    def required_headers(self) -> dict[str, Any]:
        """
        Generates the headers that must be sent to this endpoint based on its attributes

        :return:
            Header name, header value pairs
        :rtype:
            dict
        """
        return {}

    def get_formatted_path(self, **kwargs) -> str:
        """
        Format this endpoint's path with the supplied keyword arguments

        :return:
            The fully-formatted path
        :rtype:
            str
        """
        self._validate_path_placeholders(self.path_placeholders, kwargs)

        return self.path.format(**kwargs)

    @property
    def path_placeholders(self) -> list[str]:
        """
        The formattable placeholders from this endpoint's path, in the order they appear.

        Example:

            >>> endpoint = Endpoint(path='/api/{foo}/{bar}')
            >>> endpoint.path_placeholders
            ['foo', 'bar']
        """

        parser = string.Formatter()
        return [placeholder_name for _, placeholder_name, _, _ in parser.parse(self.path) if placeholder_name]

    def _validate_path_placeholders(self, placeholder_names: list[str], path_kwargs: dict[str, Any]):
        if any(path_kwarg not in placeholder_names for path_kwarg in path_kwargs):
            warnings.warn(
                f"An unknown path kwarg was supplied to {self}. kwargs supplied: {path_kwargs}",
                RuntimeWarning,
                stacklevel=6,
            )

    def _check_for_empty_params(self, params: dict[str, Any]):
        empty_params = {param: params[param] for param in params if params[param] in (None, "")}

        if empty_params:
            warnings.warn(
                f"The {self.path} endpoint was called with empty parameters: {empty_params}",
                RuntimeWarning,
                stacklevel=6,
            )

    def _check_for_unfulfilled_params(self, params: dict[str, Any]):
        unfulfilled_params = {
            param for param in self.required_params if param not in params and param not in self.default_params
        }

        if unfulfilled_params:
            raise UnfulfilledParameterException(self.path, unfulfilled_params)

    def _validate_params(self, params: dict[str, Any]):
        self._check_for_empty_params(params)
        self._check_for_unfulfilled_params(params)

    def get_merged_params(self, supplied_params: dict[str, Any] | None = None) -> dict[str, Any]:
        """
        Merge this endpoint's default parameters with the supplied parameters

        :param dict supplied_params:
            A dictionary of query parameter, value pairs
        :return:
            A dictionary of this endpoint's default parameters, merged with the supplied parameters.
            Any default parameters which have a value supplied are overridden.
        :rtype:
            dict
        :raises apiron.exceptions.UnfulfilledParameterException:
            When a required parameter for this endpoint is not a default param and is not supplied by the caller
        """
        supplied_params = supplied_params or {}

        self._validate_params(supplied_params)

        merged_params = self.default_params.copy()
        merged_params.update(supplied_params)
        return merged_params

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path='{self.path}')"



from typing import ClassVar

from apiron import Endpoint


class ServiceMeta(type):
    @property
    def required_headers(cls) -> dict[str, str]:
        return cls().required_headers

    @property
    def endpoints(cls) -> set[Endpoint]:
        return {attr for attr_name, attr in cls.__dict__.items() if isinstance(attr, Endpoint)}

    def __str__(cls) -> str:
        return str(cls())

    def __repr__(cls) -> str:
        return repr(cls())


class ServiceBase(metaclass=ServiceMeta):
    auth = ()
    proxies: ClassVar[dict[str, str]] = {}

    @property
    def required_headers(self) -> dict[str, str]:
        """
        The headers that are required to be present in requests to this service.

        :return:
            A dictionary of header names and their expected values
        :rtype:
            dict
        """
        return {}

    @classmethod
    def get_hosts(cls) -> list[str]:
        """
        The fully-qualified hostnames that correspond to this service.
        These are often determined by asking a load balancer or service discovery mechanism.

        :return:
            The hostname strings corresponding to this service
        :rtype:
            list
        """
        return []


class Service(ServiceBase):
    """
    A base class for low-level services.

    A service has a domain off of which one or more endpoints stem.
    """

    domain: str

    @classmethod
    def get_hosts(cls) -> list[str]:
        """
        The fully-qualified hostnames that correspond to this service.
        These are often determined by asking a load balancer or service discovery mechanism.

        :return:
            The hostname strings corresponding to this service
        :rtype:
            list
        """
        return [cls.domain]

    def __str__(self) -> str:
        return getattr(self.__class__, "domain", "UNKNOWN")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(domain={self.__class__.domain})"



class APIException(Exception):
    pass


class NoHostsAvailableException(APIException):
    def __init__(self, service_name: str):
        message = f"No hosts available for service: {service_name}"
        super().__init__(message)


class UnfulfilledParameterException(APIException):
    def __init__(self, endpoint_path: str, unfulfilled_params: set[str]):
        message = f"The {endpoint_path} endpoint was called without required parameters: {unfulfilled_params}"
        super().__init__(message)
